from colt import Colt
from colt.colt import ColtMeta
from abc import abstractmethod
import subprocess as sp
from .atominfo import atomnumber_to_atomname
from .gauwrite import write_gaufile


def write_content_to_file(fileName, content, options="w"):
    """ write content to file [fileName]

        fileName = str, Name of the file to be read
        content  = str, content written to the file
    """
    with open(fileName, options) as f:
        f.write(content)


class Logger:
    """Basic Logger"""

    def __init__(self, filename):
        self.fh = open(filename, 'w')

    def __del__(self):
        if self.fh.closed is False:
            self.fh.close()

    def write(self, msg):
        self.fh.write(f"{msg}\n")

    def warning(self, msg):
        self.fh.write(f"Warning: {msg}\n")

    def error(self, msg, error='ValueError'):
        self.fh.write(f"{error}: {msg}\n")
        self.fh.close()
        if error == 'ErrorTermination':
            error = 'SystemExit'

        if error == 'ValueError':
            raise ValueError(msg)
        if error == 'SystemExit':
            raise SystemExit(msg)
        raise SystemExit


_GAUSSIAN_BASE_INTERFACE = """
# which layer in a ONIOM Calculation
layer = :: str :: R, M, S
# Name of the Gau-xxx.EIn file
inputfile = :: existing_file
# Name of the Gau-xxx.EOu file
outputfile = :: file
# Write additional information to gauout
msgfile = :: file
# name of a gaussian fchk file
fchkfile = :: file
# 
matelfile = :: file
"""

class ExternalMeta(ColtMeta):

    def __new__(cls, name, bases, clsdict):
        if '_user_input' in clsdict:
            clsdict['_user_input'] += _GAUSSIAN_BASE_INTERFACE
        else:
            clsdict['_user_input'] = _GAUSSIAN_BASE_INTERFACE

        return ColtMeta.__new__(cls, name, bases, clsdict)


class GaussianExternalInterface(Colt, metaclass=ExternalMeta):
    """ Read Gaussian External Input File """

    au2ang = 1.0/1.889725949297217

    def __init__(self, input_file):
        """ Sets up GaussianExternalInterface

            :
        """
        self.inp_file = input_file

    @property
    def method(self):
        return self.inp_file.imethod

    @property
    def logger(self):
        return self.inp_file.logger

    def log_write(self, msg):
        return self.inp_file.logger.write(msg)

    def log_warning(self, msg):
        return self.inp_file.logger.warning(msg)

    def log_error(self, msg, error='SystemExit'):
        return self.inp_file.logger.error(msg, error=error)

    @property
    def chg(self):
        return self.inp_file.chg

    @property
    def natoms(self):
        return self.inp_file.natoms

    @property
    def atomids(self):
        return self.inp_file.atomids

    @property
    def charges(self):
        return self.inp_file.charges

    @property
    def coords(self):
        return self.inp_file.coords

    @classmethod
    def run(cls):
        self = cls.from_commandline()
        self.do_calculation()

    def do_calculation(self):
        # write the inputs that are method dependent and run the corresponding code
        (e, d), grad, hess = self.__run_interface()
        self.inp_file.save(e, d, gradient=grad, force_constants=hess)

    def __run_interface(self):
        if self.method == 0:
            result = self._run_energy()
        elif self.method == 1:
            result = self._run_gradient()
        elif self.method == 2:
            result = self._run_frequency()
        else:
            raise ValueError(("Input Method can only be '0, 1, 2' ",
                              "but not %d" % self.method))
        return result

    @abstractmethod
    def _run_energy(self):
        ...

    @abstractmethod
    def _run_gradient(self):
        ...

    @abstractmethod
    def _run_frequency(self):
        ...

    def run_interface(self, script, name):
        self.log_write(f"""
 ############## Run section: GAUEXT  ##############
    Interface {name} call:
 --------------------------------------------------
    {script}
 --------------------------------------------------
 """)
        error = sp.call(script, shell=True)
        if error == 0:
            self.log_write("Normal Termination of interface")
            return True
        self.log_write('Error Termination: *** Something went wrong! ***')
        self.log_write("Error count: {error}")
        return False

    @classmethod
    def write_xyz_file(cls, filename, natoms, comment, atomids, coords):
        txt = "".join(["%d\n" % natoms,
                       "%s\n" % comment,
                       "".join(cls.get_coordinates(atomid, coord)
                               for atomid, coord in zip(atomids, coords))
                       ])
        write_content_to_file(filename, txt)

    @classmethod
    def get_coordinates(cls, atomid, coords):
        return "% 4s    % 14.10f   % 14.10f   % 14.10f \n" % (
               atomid, cls.au2ang*coords[0], cls.au2ang*coords[1], cls.au2ang*coords[2])



class GaussianExternalInput:

    @classmethod
    def from_config(cls, config):
        return cls(config['inputfile'], 
                   config['msgfile'],
                   config['layer'], 
                   config['outputfile'],
                   config['fchkfile'],
                   config['matelfile'])

    def __init__(self, inputfile, msgfile, layer=None, outputfile=None, fchkfile=None, matelfile=None):
        """
        > gauinp = ExternalInput(filename)
        > coords = gauinp.coords
        > NAtoms = gauinp.natoms

        :str filename: Gau-XXX.EIn, name of the external input file
                       created by gaussian

        """
        self.inputfile = inputfile
        self.outputfile = outputfile
        self.fchkfile = fchkfile
        self.matelfile = matelfile
        #
        self.logger = Logger(msgfile)
        #
        (self.natoms, self.imethod, self.chg,
            self.mult, self.coords, self.atomids, self.charges, self.conn) = self.parse(inputfile)


    def save(self, energy, dipole, gradient=None, polarizability=None, dipole_deriv=None, force_constants=None):
        self.logger.write("Saving External Interface results")
        write_gaufile(self.outputfile, self.imethod, self.natoms, energy, dipole, gradient=gradient, force_constants=force_constants)
        self.logger.write("Exiting Interface")

    def _get_coordinates(self, line):
        res = line.split()
        return (atomnumber_to_atomname[int(res[0])],
                [float(res[1]), float(res[2]), float(res[3])],
                float(res[4]))

    def _get_connectivity(self, line):
        """ getConnectivity as non redundant gaussian connectivity """

        columns = line.split()
        if len(columns) == 1.0:
            return []
        return [int(i) for j, i in columns[1:] if j % 2 == 0]

    def _get_redundant_connectivity(self, conn):
        """ convert non redundant connectivity in reduant one """
        for i, ele in enumerate(conn):
            for e in ele:
                if i+1 not in conn[e-1]:
                    conn[e-1].append(i+1)
        return conn

    def parse(self, filename, conn=False, tConn='red'):
        """ Parse the gaussian external input file """
        with open(filename, "r") as f:
            line = f.readline()
            natoms, imethod, chg, mult = map(int, line.split())
            coords = []
            atomids = []
            charges = []
            for _ in range(natoms):
                atomid, coord, charge = self._get_coordinates(f.readline())
                coords.append(coord)
                atomids.append(atomid)
                charges.append(charge)
            if conn is True:
                connectivity = tuple(self._get_connectivity(line) for line in f.readline())
                if tConn == 'red':
                    connectivity = self._get_redundant_connectivity(connectivity)
            else:
                connectivity = None
        self.logger.write(f"Reading file '{filename}' completed: NAtoms={natoms}, Method={imethod}, chg={chg}, mult={mult}") 
        return natoms, imethod, chg, mult, coords, atomids, charges, connectivity

    def _write_coordinate_line(self, coord):
        return " %8s  %12.8f  %12.8f %12.8f " % (coord[0], coord[1][0], coord[1][1], coord[1][2])



import os
from colt import Colt
import numpy as np

from .gaussian import GaussianExternalInterface, GaussianExternalInput





class XTBInterface(GaussianExternalInterface):

    name = "XTB"
    executable = 'xtb'
    _command = "%s %s --chrg %d %s --wbo "

    _user_input = """
    options = :: str, optional, alias=o
    save = false :: bool, alias=s
    input = xtbinput.xyz :: file
    """

    @classmethod
    def from_config(cls, config):
        inp = GaussianExternalInput.from_config(config)
        return cls(inp, config['input'], config['options'], config['save'])

    def __init__(self, inp, xtbfilename, options, save):
        super().__init__(inp)
        #
        if options is None:
            options = " "
        self._options = options
        self._save = save
        self.input = "xtb.xyz"

    def _write_general_inputs(self):
        self.write_xyz_file(self.input, self.inp_file.natoms, "xtb input", self.inp_file.atomids, self.inp_file.coords)

    def set_options(self, options, save):
        self._save = save
        if options is None:
            return
        self._options = options

    def _get_command(self, options):
        return self._command % (self.executable, self.input, self.chg, self._options + options)

    def _handle_xcontrol(self):
        if self._save is True:
            return
        if os.path.isfile("xcontrol"):
           os.remove("xcontrol")

    def _run_interface(self, command):
        self._handle_xcontrol()
        self._write_general_inputs()
        self.run_interface(command, self.name)

    def _run_energy(self):
        command = self._get_command(" --sp --copy > output_energy ")
        self._run_interface(command)
        self.logger.write(get_gaussian_wbo_txt(self.inp_file.atomids, 'wbo'))
        self.logger.write(get_gaussian_charges(self.inp_file.atomids, 'charges'))
        return (self._read_xtb_energy_and_dipole("output_energy"),
                None, None)

    def _run_gradient(self):
        command = self._get_command(" --grad --copy > output_gradient ")
        self._run_interface(command)
        self.logger.write(get_gaussian_wbo_txt(self.inp_file.atomids, 'wbo'))
        self.logger.write(get_gaussian_charges(self.inp_file.atomids, 'charges'))
        output = (self._read_xtb_energy_and_dipole("output_gradient"),
                self._read_xtb_gradient("gradient"), None)
        self.clean_up()
        return output

    def _run_frequency(self):
        e_and_dip, grad, _ = self._run_gradient()
        command = self._get_command(" --hess --copy > output_freq")
        self._run_interface(command)
        return (e_and_dip,
                grad,
                self._read_xtb_hessian("hessian"))

    def clean_up(self):
        os.rename("gradient", "grad_old")
        return

    def _read_xtb_energy_and_dipole(self, filename):
        with open(filename, 'r') as fh:
            for line in fh:
                if line.startswith('molecular dipole:'):
                    dipole = _parse_diple(fh)
                if 'TOTAL ENERGY' in line:
                    energy = float(line.split()[3])
                    return energy, dipole

    def _read_xtb_gradient(self, filename):
        with open(filename, 'r') as fh:
            itr = iter(fh)
            for _ in range(self.natoms+2):
                next(itr)
            return np.array([[float(val) for val in next(itr).split()]
                              for _ in range(self.natoms)])

    def _read_xtb_hessian(self, filename):
        with open(filename, 'r') as fh:
            itr = iter(fh)
            next(itr)
            return np.array([float(entry) 
                             for line in itr 
                             for entry in line.split()])


def iformat(num):
    return str(num).center(6)


def get_header(start, end):
    first =  "     Atom"
    second = "     ----"
    for i in range(start+1, end+1):
        first += " " + iformat(i) + " "
        second += " ------ "
    return first + "\n" +  second + "\n"


def get_wbo_line(atom, wbo, idx, istart, end):
    line = str(idx+1)
    line = " "*(4-len(line)) + line + '.  ' + atom + " "*(2-len(atom))
    for i in range(istart, end):
        line += "%7.4f " % wbo[idx, i]
    return line + '\n'


def get_block(atomids, wbo, istart, end):
    txt = get_header(istart, end) 
    for i, atom in enumerate(atomids):
        txt += get_wbo_line(atom, wbo, i, istart, end)
    return txt + "\n"

    
def get_gaussian_wbo_txt(atomids, filename):
    txt = """

-------------------------------------------------------------
    Start of: N A T U R A L   B O N D   O R B I T A L
-------------------------------------------------------------

 Wiberg bond index matrix in the NAO basis:

 """
    natoms = len(atomids)
    wbo = parse_wbo(natoms, filename)
    istart = 0
    while (istart < natoms):
        end = istart + 9
        if end >= natoms:
            end = natoms
        txt += get_block(atomids, wbo, istart, end)
        istart = end
    txt += """   
 Wiberg bond index, Totals by atom:                                            

 """
    wbosum = np.zeros((natoms, 1))
    for i in range(natoms):
        wbosum[i, 0] = sum(wbo[i])
    txt += get_block(atomids, wbosum, 0, 1)
    txt += "\n\n --- END WBO ANALYSIS ---\n\n"
    return txt


def parse_wbo(natoms, filename):
    wbo = np.zeros((natoms, natoms))
    with open(filename, 'r') as fh:
        for line in fh:
            i, j, value = line.split()
            i, j, value = int(i)-1, int(j)-1, float(value)
            wbo[i,j] = value
            wbo[j,i] = value
    return wbo


def _parse_charges(filename):
    with open(filename, 'r') as fh:
        return [float(val) for val in fh]

def get_gaussian_charges(atomids, filename):
    charges =  _parse_charges(filename)
    assert len(atomids) == len(charges), "inconsistency between charges and natoms"
    charge_txt = "\n".join("%12.8f" % charge for charge in charges)
    return f"Hirshfeld charges, spin densities\n{charge_txt}\n\n\nESP charges:\n{charge_txt}"


def _parse_diple(fh):
    for _ in range(2):
        next(fh)
    cols = next(fh).split()
    return [float(cols[1]), float(cols[2]), float(cols[3])]



from abc import ABC, abstractmethod
import subprocess as sp
#
from .fileio import write_content_to_file




class GaussianExternalInterface(ABC):

    au2ang = 1.0/1.889725949297217

    def __init__(self, input_file): 
        """ Sets up GaussianExternalInterface

            :
        """
        self.inp_file = input_file

    @property
    def method(self):
        return self.inp_file.imethod

    @property
    def logger(self):
        return self.inp_file.logger

    def log_write(self, msg):
        return self.inp_file.logger.write(msg)

    def log_warning(self, msg):
        return self.inp_file.logger.warning(msg)

    def log_error(self, msg, error='SystemExit'):
        return self.inp_file.logger.error(msg, error=error)

    @property
    def chg(self):
        return self.inp_file.chg

    @property
    def natoms(self):
        return self.inp_file.natoms

    @property
    def atomids(self):
        return self.inp_file.atomids

    @property
    def charges(self):
        return self.inp_file.charges

    @property
    def coords(self):
        return self.inp_file.coords

    def do_calculation(self):
        # write the inputs that are method dependent and run the corresponding code
        return self.__run_interface()

    def __run_interface(self):
        if self.method == 0:
            result = self._run_energy()
        elif self.method == 1:
            result = self._run_gradient()
        elif self.method == 2:
            result = self._run_frequency()
        else:
            raise ValueError(("Input Method can only be '0, 1, 2' ",
                              "but not %d" % self.method))
        return result

    @abstractmethod
    def _run_energy(self):
        ...

    @abstractmethod
    def _run_gradient(self):
        ...

    @abstractmethod
    def _run_frequency(self):
        ...

    def run_interface(self, script, name):
        self.log_write("""
 ############## Run section: GAUEXT  ##############
    Interface {name} call:
 --------------------------------------------------
    {script}
 --------------------------------------------------
 """)
        error = sp.call(script, shell=True)
        if error == 0:
            self.log_write("Normal Termination of interface")
            return True
        self.log_write('Error Termination: *** Something went wrong! ***')
        self.log_write("Error count: {error}")
        return False

    @classmethod
    def write_xyz_file(cls, filename, natoms, comment, atomids, coords):
        txt = "".join(["%d\n" % natoms,
                       "%s\n" % comment,
                       "".join(cls.get_coordinates(atomid, coord)
                               for atomid, coord in zip(atomids, coords))
                       ])
        write_content_to_file(filename, txt)

    @classmethod
    def get_coordinates(cls, atomid, coords):
        return "% 4s    % 14.10f   % 14.10f   % 14.10f \n" % (
               atomid, cls.au2ang*coords[0], cls.au2ang*coords[1], cls.au2ang*coords[2])



from libc.stdio cimport sprintf, fprintf, FILE, fopen, fclose
cimport numpy as np
import numpy as np


cdef void fortran_format_i(const int f, char* string):
    sprintf(string, "%20.12e", f);
    string[16] = b'D';

cdef void fortran_format_d(const double f, char* string):
    sprintf(string, "%20.12e", f);
    string[16] = b'D';

cdef void fortran_format_f(const float f, char* string):
    sprintf(string, "%20.12e", f);
    string[16] = b'D';

cdef void fprintout(FILE* fp, np.ndarray f, const int N):
    assert f.dtype == np.double or f.dtype == np.int

    cdef char[21] string;

    if f.dtype == np.double:
        for i, ele in enumerate(f):
            fortran_format_d(ele, string)
            fprintf(fp, string);
            if i % N == N-1:
                fprintf(fp, "\n");
    elif f.dtype == np.int:
        for i, ele in enumerate(f):
            fortran_format_i(ele, string);
            fprintf(fp, string);
            if i % N == N-1:
                fprintf(fp, "\n");

cdef void display_zeros(FILE* fp, const int nele):
    cdef np.ndarray value = np.zeros((nele,), dtype=np.double);
    fprintout(fp, value, 3)


cdef np.ndarray get_forceconstants(np.ndarray force, const int natoms):
    cdef int n3n3 = natoms*3*natoms*3
    cdef int nhalf = natoms*3*(3*natoms+1)//2
    cdef np.ndarray output
    cdef int counter
    cdef int s

    if force.size == nhalf:
        return force

    if force.size != n3n3:
        raise ValueError("force constant matrix can only be matrix of lower triangular")

    output = np.empty(nhalf, dtype=np.double)
    counter = 0
    for i in range(3*natoms):
        s = i*3*natoms
        for j in range(i+1):
            output[counter] = force[s + j]
            counter += 1
    return output
             

def write_gaufile(filename, iopts, natoms, energy, dipole,
                  gradient=None, polarizability=None, dipole_deriv=None,
                  force_constants=None):
    cdef FILE* fp;
    cdef char[21] string;
    cdef np.ndarray _value;
    cdef np.ndarray e_and_dip = np.array([energy, dipole[0], dipole[1], dipole[2]], dtype=np.double)
    cdef np.ndarray lowertriangular
    fp = fopen(filename.encode('utf-8'), "w")
    # write e_and_dip
    fprintout(fp, e_and_dip, 4)
    if iopts == 0:
        pass
    else:
        #
        if gradient is None:
            raise ValueError("Gradient asked, but not provided")
        fprintout(fp, gradient.flatten(), 3)
            
        if iopts == 2:
            if force_constants is None:
                raise ValueError("Frequency asked but not provided")
            #

            if polarizability is None:
                display_zeros(fp, 6)
            else:
                fprintout(fp, polarizability.flatten(), 3)

            if dipole_deriv is None:
                display_zeros(fp, 9*natoms)
            else:
                fprintout(fp, dipole_deriv.flatten(), 3)
            # no consistency check...
            lowertriangular = get_forceconstants(force_constants, natoms)
            fprintout(fp, lowertriangular, 3)
    #
    fclose(fp);



from setuptools import setup, find_packages, Extension
import os

import numpy as np
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None


with open('README.rst') as readme_file:
    readme = readme_file.read()


requirements = ['cython', 'numpy', 'pycolt']


# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


extensions = [Extension("gauext.gauwrite", ["gauext/gauwrite.pyx"],
                        include_dirs=[np.get_include()]),
                        ]

CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 0))) and cythonize is not None


if CYTHONIZE:
    compiler_directives = {"language_level": 3, "embedsignature": True}
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)


setup(
    author="Maximilian F.S.J. Menger",
    author_email='m.f.s.j.menger@rug.nl',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    license="Apache License v2.0",
    description="Generic Gaussian External Interface",
    install_requires=requirements,
    ext_modules = extensions,
    packages=find_packages(include=['gauext', 'gauext.*']),
    keywords='Gaussian, Chemistry',
    package_data = {
        'gauext': ['*.pyx'],
    },
    entry_points = {
        'console_scripts': ['gauext-xtb=gauext.xtb:XTBInterface.run', ],
    },
    long_description=readme,
    name='gauext',
    url='https://github.com/mfsjmenger/gauext',
    version='0.2.1',
    zip_safe=False,
)
